#ifndef VOCAL_TRACT_MODEL_PARAMETER_VALUE_H
#define VOCAL_TRACT_MODEL_PARAMETER_VALUE_H

namespace GS {

struct VocalTractModelParameterValue {
	int index;
	double value;
};

} /* namespace GS */

#endif // VOCAL_TRACT_MODEL_PARAMETER_VALUE_H
